@@ -1,27 +1,32 @@
+// SPDX-License-Identifier: Apache-2.0 OR MIT
+
 //! Cargo configuration that environment variables, config overrides, and
 //! target-specific configurations have not been resolved.
 
 #[path = "gen/de.rs"]
 mod gen;
 
+use core::{fmt, slice, str::FromStr};
 use std::{
     borrow::Cow,
     collections::BTreeMap,
     ffi::OsStr,
-    fmt::{self, Formatter},
     fs,
     path::{Path, PathBuf},
-    slice,
-    str::FromStr,
 };
 
-use serde::{Deserialize, Serialize};
+use serde::{
+    de::{self, Deserialize, Deserializer},
+    ser::{Serialize, Serializer},
+};
+use serde_derive::{Deserialize, Serialize};
 
 pub use crate::value::{Definition, Value};
 use crate::{
     easy,
     error::{Context as _, Error, Result},
     resolve::{ResolveContext, TargetTripleRef},
+    walk,
 };
 
 /// Cargo configuration that environment variables, config overrides, and
@@ -106,15 +111,15 @@ impl Config {
     }
 
     /// Read config files hierarchically from the given directory and merges them.
-    pub fn load_with_cwd(cwd: impl AsRef<Path>) -> Result<Self> {
+    pub fn load_with_cwd<P: AsRef<Path>>(cwd: P) -> Result<Self> {
         let cwd = cwd.as_ref();
-        Self::_load_with_options(cwd, home::cargo_home_with_cwd(cwd).ok())
+        Self::_load_with_options(cwd, walk::cargo_home_with_cwd(cwd))
     }
 
     /// Read config files hierarchically from the given directory and merges them.
-    pub fn load_with_options(
-        cwd: impl AsRef<Path>,
-        cargo_home: impl Into<Option<PathBuf>>,
+    pub fn load_with_options<P: AsRef<Path>, Q: Into<Option<PathBuf>>>(
+        cwd: P,
+        cargo_home: Q,
     ) -> Result<Self> {
         Self::_load_with_options(cwd.as_ref(), cargo_home.into())
     }
@@ -143,13 +148,13 @@ impl Config {
     ///
     /// **Note:** Note: This just reads a file at the given path and does not
     /// respect the hierarchical structure of the cargo config.
-    pub fn load_file(path: impl AsRef<Path>) -> Result<Self> {
+    pub fn load_file<P: AsRef<Path>>(path: P) -> Result<Self> {
         Self::_load_file(path.as_ref())
     }
     fn _load_file(path: &Path) -> Result<Self> {
         let buf = fs::read_to_string(path)
             .with_context(|| format!("failed to read `{}`", path.display()))?;
-        let mut config: Config = toml::from_str(&buf).with_context(|| {
+        let mut config: Config = toml_edit::de::from_str(&buf).with_context(|| {
             format!("failed to parse `{}` as cargo configuration", path.display())
         })?;
         config.set_path(path);
@@ -214,6 +219,12 @@ impl Config {
                 continue;
             }
             if cx.eval_cfg(k, target_triple, build_config)? {
+                // https://github.com/rust-lang/cargo/pull/12535
+                if target_linker.is_none() {
+                    if let Some(linker) = v.linker.as_ref() {
+                        target_linker = Some(linker.clone());
+                    }
+                }
                 // Priorities (as of 1.68.0-nightly (2022-12-23)):
                 // 1. CARGO_TARGET_<triple>_RUNNER
                 // 2. target.<triple>.runner
@@ -482,7 +493,7 @@ pub struct RegistriesConfigValue {
 }
 
 impl fmt::Debug for RegistriesConfigValue {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let Self { index, token, protocol } = self;
         let redacted_token = token
             .as_ref()
@@ -549,7 +560,7 @@ pub struct RegistryConfig {
 }
 
 impl fmt::Debug for RegistryConfig {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let Self { default, token } = self;
         let redacted_token = token
             .as_ref()
@@ -788,7 +799,7 @@ impl Flags {
 impl<'de> Deserialize<'de> for Flags {
     fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
     where
-        D: serde::Deserializer<'de>,
+        D: Deserializer<'de>,
     {
         let v: StringOrArray = Deserialize::deserialize(deserializer)?;
         match v {
@@ -874,7 +885,7 @@ impl PathAndArgs {
 impl Serialize for PathAndArgs {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
-        S: serde::Serializer,
+        S: Serializer,
     {
         match self.deserialized_repr {
             StringListDeserializedRepr::String => {
@@ -899,9 +910,8 @@ impl Serialize for PathAndArgs {
 impl<'de> Deserialize<'de> for PathAndArgs {
     fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
     where
-        D: serde::Deserializer<'de>,
+        D: Deserializer<'de>,
     {
-        use serde::de::Error;
         #[derive(Deserialize)]
         #[serde(untagged)]
         enum StringOrArray {
@@ -915,7 +925,7 @@ impl<'de> Deserialize<'de> for PathAndArgs {
         };
         match res {
             Some(path) => Ok(path),
-            None => Err(D::Error::invalid_length(0, &"at least one element")),
+            None => Err(de::Error::invalid_length(0, &"at least one element")),
         }
     }
 }
@@ -960,7 +970,7 @@ impl StringList {
 impl Serialize for StringList {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
-        S: serde::Serializer,
+        S: Serializer,
     {
         match self.deserialized_repr {
             StringListDeserializedRepr::String => {
@@ -983,7 +993,7 @@ impl Serialize for StringList {
 impl<'de> Deserialize<'de> for StringList {
     fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
     where
-        D: serde::Deserializer<'de>,
+        D: Deserializer<'de>,
     {
         let v: StringOrArray = Deserialize::deserialize(deserializer)?;
         match v {
