@@ -1,15 +1,16 @@
+// SPDX-License-Identifier: Apache-2.0 OR MIT
+
+use core::{cell::RefCell, fmt, ops};
 use std::{
     borrow::Cow,
-    cell::RefCell,
     collections::BTreeMap,
     ffi::{OsStr, OsString},
-    fmt::{self, Formatter},
-    ops,
     path::{Path, PathBuf},
     process::Command,
 };
 
-use serde::Serialize;
+use serde::ser::{Serialize, Serializer};
+use serde_derive::Serialize;
 
 use crate::{
     de::{self, split_encoded, split_space_separated, Color, Frequency, RegistriesProtocol, When},
@@ -99,8 +100,6 @@ pub struct Config {
     // Resolve contexts. Completely ignored in serialization and deserialization.
     #[serde(skip)]
     cx: ResolveContext,
-    #[serde(skip)]
-    current_dir: PathBuf,
 }
 
 fn ref_cell_bree_map_is_empty<K, V>(map: &RefCell<BTreeMap<K, V>>) -> bool {
@@ -114,36 +113,32 @@ impl Config {
     }
 
     /// Read config files hierarchically from the given directory and merges them.
-    pub fn load_with_cwd(cwd: impl AsRef<Path>) -> Result<Self> {
+    pub fn load_with_cwd<P: AsRef<Path>>(cwd: P) -> Result<Self> {
         let cwd = cwd.as_ref();
         Self::load_with_options(cwd, ResolveOptions::default())
     }
 
     /// Read config files hierarchically from the given directory and merges them.
-    pub fn load_with_options(cwd: impl AsRef<Path>, options: ResolveOptions) -> Result<Self> {
+    pub fn load_with_options<P: AsRef<Path>>(cwd: P, options: ResolveOptions) -> Result<Self> {
         let cwd = cwd.as_ref();
-        let cx = options.into_context();
+        let cx = options.into_context(cwd.to_owned());
 
-        let de = de::Config::_load_with_options(cwd, cx.cargo_home(cwd).clone())?;
-        Self::from_unresolved(de, cx, cwd.to_owned())
+        let de = de::Config::_load_with_options(&cx.current_dir, cx.cargo_home(cwd).clone())?;
+        Self::from_unresolved(de, cx)
     }
 
-    pub(crate) fn from_unresolved(
-        mut de: de::Config,
-        cx: ResolveContext,
-        current_dir: PathBuf,
-    ) -> Result<Self> {
+    pub(crate) fn from_unresolved(mut de: de::Config, cx: ResolveContext) -> Result<Self> {
         de.apply_env(&cx)?;
 
         let mut alias = BTreeMap::new();
         for (k, v) in de.alias {
             alias.insert(k, StringList::from_unresolved(v));
         }
-        let build = BuildConfig::from_unresolved(de.build, &current_dir);
-        let doc = DocConfig::from_unresolved(de.doc, &current_dir);
+        let build = BuildConfig::from_unresolved(de.build, &cx.current_dir);
+        let doc = DocConfig::from_unresolved(de.doc, &cx.current_dir);
         let mut env = BTreeMap::new();
         for (k, v) in de.env {
-            env.insert(k, EnvConfigValue::from_unresolved(v, &current_dir));
+            env.insert(k, EnvConfigValue::from_unresolved(v, &cx.current_dir));
         }
         let future_incompat_report =
             FutureIncompatReportConfig::from_unresolved(de.future_incompat_report);
@@ -168,7 +163,6 @@ impl Config {
             de_target: de.target,
             term,
             cx,
-            current_dir,
         })
     }
 
@@ -252,10 +246,9 @@ impl Config {
     /// }
     /// # Ok(()) }
     /// ```
-    #[allow(single_use_lifetimes)]
-    pub fn build_target_for_config<'a>(
+    pub fn build_target_for_config<'a, I: IntoIterator<Item = T>, T: Into<TargetTripleRef<'a>>>(
         &self,
-        targets: impl IntoIterator<Item = impl Into<TargetTripleRef<'a>>>,
+        targets: I,
     ) -> Result<Vec<TargetTriple>> {
         let targets: Vec<_> = targets.into_iter().map(|v| v.into().into_owned()).collect();
         if !targets.is_empty() {
@@ -265,7 +258,7 @@ impl Config {
         if !config_targets.is_empty() {
             return Ok(config_targets);
         }
-        Ok(vec![TargetTripleRef::from(self.cx.host_triple()?).into_owned()])
+        Ok(vec![TargetTripleRef::from(self.cx.host_triple(&self.build)?).into_owned()])
     }
 
     /// Selects target triples to pass to CLI.
@@ -283,10 +276,10 @@ impl Config {
     /// above) is set.
     /// Also, Unlike [`build_target_for_config`](Self::build_target_for_config)
     /// the target name specified in path is preserved.
-    #[allow(clippy::unnecessary_wraps)]
-    pub fn build_target_for_cli(
+    #[allow(clippy::unnecessary_wraps)] // TODO: change in next breaking release?
+    pub fn build_target_for_cli<I: IntoIterator<Item = S>, S: AsRef<str>>(
         &self,
-        targets: impl IntoIterator<Item = impl AsRef<str>>,
+        targets: I,
     ) -> Result<Vec<String>> {
         let targets: Vec<_> = targets.into_iter().map(|t| t.as_ref().to_owned()).collect();
         if !targets.is_empty() {
@@ -312,39 +305,35 @@ impl Config {
                     &self.build,
                 )?
                 .unwrap_or_default(),
-                &self.current_dir,
+                &self.cx.current_dir,
             );
             target_configs.insert(TargetTripleBorrow(target.clone().into_owned()), target_config);
         }
         Ok(())
     }
     /// Returns the resolved `[target]` table for the given target.
-    #[allow(single_use_lifetimes)]
-    pub fn target<'a>(&self, target: impl Into<TargetTripleRef<'a>>) -> Result<TargetConfig> {
+    pub fn target<'a, T: Into<TargetTripleRef<'a>>>(&self, target: T) -> Result<TargetConfig> {
         let target = target.into();
         self.init_target_config(&target)?;
         Ok(self.target.borrow()[target.cli_target()].clone())
     }
     /// Returns the resolved linker path for the given target.
-    #[allow(single_use_lifetimes)]
-    pub fn linker<'a>(&self, target: impl Into<TargetTripleRef<'a>>) -> Result<Option<PathBuf>> {
+    pub fn linker<'a, T: Into<TargetTripleRef<'a>>>(&self, target: T) -> Result<Option<PathBuf>> {
         let target = target.into();
         self.init_target_config(&target)?;
         Ok(self.target.borrow()[target.cli_target()].linker.clone())
     }
     /// Returns the resolved runner path and args for the given target.
-    #[allow(single_use_lifetimes)]
-    pub fn runner<'a>(
+    pub fn runner<'a, T: Into<TargetTripleRef<'a>>>(
         &self,
-        target: impl Into<TargetTripleRef<'a>>,
+        target: T,
     ) -> Result<Option<PathAndArgs>> {
         let target = target.into();
         self.init_target_config(&target)?;
         Ok(self.target.borrow()[target.cli_target()].runner.clone())
     }
     /// Returns the resolved rustflags for the given target.
-    #[allow(single_use_lifetimes)]
-    pub fn rustflags<'a>(&self, target: impl Into<TargetTripleRef<'a>>) -> Result<Option<Flags>> {
+    pub fn rustflags<'a, T: Into<TargetTripleRef<'a>>>(&self, target: T) -> Result<Option<Flags>> {
         let target = target.into();
         self.init_target_config(&target)?;
         Ok(self.target.borrow()[target.cli_target()].rustflags.clone())
@@ -371,7 +360,7 @@ impl Config {
     }
     /// Returns the host triple.
     pub fn host_triple(&self) -> Result<&str> {
-        self.cx.host_triple()
+        self.cx.host_triple(&self.build)
     }
 
     // TODO: add override instead?
@@ -632,7 +621,7 @@ impl EnvConfigValue {
 impl Serialize for EnvConfigValue {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
-        S: serde::Serializer,
+        S: Serializer,
     {
         #[derive(Serialize)]
         #[serde(untagged)]
@@ -761,7 +750,7 @@ impl RegistriesConfigValue {
 }
 
 impl fmt::Debug for RegistriesConfigValue {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let Self { index, token, protocol } = self;
         let redacted_token = token.as_ref().map(|_| "[REDACTED]");
         f.debug_struct("RegistriesConfigValue")
@@ -807,7 +796,7 @@ impl RegistryConfig {
 }
 
 impl fmt::Debug for RegistryConfig {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let Self { default, token } = self;
         let redacted_token = token.as_ref().map(|_| "[REDACTED]");
         f.debug_struct("RegistryConfig")
@@ -992,7 +981,7 @@ impl Flags {
     }
 
     /// Appends a flag to the back of this rustflags.
-    pub fn push(&mut self, flag: impl Into<String>) {
+    pub fn push<S: Into<String>>(&mut self, flag: S) {
         self.flags.push(flag.into());
     }
 }
@@ -1035,16 +1024,16 @@ pub struct PathAndArgs {
 
 impl PathAndArgs {
     /// Creates a new program.
-    pub fn new(path: impl Into<PathBuf>) -> Self {
+    pub fn new<P: Into<PathBuf>>(path: P) -> Self {
         Self { path: path.into(), args: vec![] }
     }
     /// Adds an argument to pass to the program.
-    pub fn arg(&mut self, arg: impl Into<OsString>) -> &mut Self {
+    pub fn arg<S: Into<OsString>>(&mut self, arg: S) -> &mut Self {
         self.args.push(arg.into());
         self
     }
     /// Adds multiple arguments to pass to the program.
-    pub fn args(&mut self, args: impl IntoIterator<Item = impl Into<OsString>>) -> &mut Self {
+    pub fn args<I: IntoIterator<Item = S>, S: Into<OsString>>(&mut self, args: I) -> &mut Self {
         self.args.extend(args.into_iter().map(Into::into));
         self
     }
@@ -1053,7 +1042,7 @@ impl PathAndArgs {
 impl Serialize for PathAndArgs {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
-        S: serde::Serializer,
+        S: Serializer,
     {
         let mut v = Vec::with_capacity(1 + self.args.len());
         v.push(self.path.to_string_lossy().into_owned());
@@ -1073,9 +1062,7 @@ impl From<PathAndArgs> for Command {
 }
 impl From<PathAndArgs> for ProcessBuilder {
     fn from(value: PathAndArgs) -> Self {
-        let mut cmd = ProcessBuilder::new(value.path);
-        cmd.args(value.args);
-        cmd
+        ProcessBuilder::from_std(Command::from(value))
     }
 }
 
